/************************************************************************
 === TinyAdafruit_PCD8544 by Tyler Bletsch (Tyler.Bletsch@gmail.com) ===
  
  This is a drastically simplified version of the Adafruit library for 
  Monochrome Nokia 5110 LCD Displays designed to work on ATTINY 
  microcontrollers with as little as 4kB of flash and 512 bytes of RAM.
 
  It's configured like the original Adafruit library, but it is  
  TEXT ONLY, substituting a text buffer for a full frame buffer.
  
  Interface-wise, it implements the Arduino standard print() and
  println() functions.
  
*************************************************************************/

/*********************************************************************
This is a library for our Monochrome Nokia 5110 LCD Displays

  Pick one up today in the adafruit shop!
  ------> http://www.adafruit.com/products/338

These displays use SPI to communicate, 4 or 5 pins are required to  
interface

Adafruit invests time and resources providing this open source code, 
please support Adafruit and open-source hardware by purchasing 
products from Adafruit!

Written by Limor Fried/Ladyada  for Adafruit Industries.  
BSD license, check license.txt for more information
All text above, and the splash screen below must be included in any redistribution
*********************************************************************/

#include <avr/pgmspace.h>
#if defined(ARDUINO) && ARDUINO >= 100
  #include "Arduino.h"
#else
  #include "WProgram.h"
#endif
#include <util/delay.h>
#include <stdlib.h>

#include "TinyAdafruit_PCD8544.h"

#include "glcdfont.c"
extern unsigned char PROGMEM font5x7[];

TinyAdafruit_PCD8544::TinyAdafruit_PCD8544(int8_t SCLK, int8_t DIN, int8_t DC, int8_t CS, int8_t RST) {
	_din = DIN;
	_sclk = SCLK;
	_dc = DC;
	_rst = RST;
	_cs = CS;
	
	wrap_method = SCROLL;
	
	clearDisplay();
}

TinyAdafruit_PCD8544::TinyAdafruit_PCD8544(int8_t SCLK, int8_t DIN, int8_t DC, int8_t RST) {
	TinyAdafruit_PCD8544(SCLK,DIN,DC,-1,RST);
}

#define rowcol2mem(r,c) ((r)*LCDTEXTWIDTH + (c))

size_t TinyAdafruit_PCD8544::write(uint8_t c) {
	uint8_t b = rowcol2mem(row,col);
	if (b >= LCDTEXTWIDTH*LCDTEXTHEIGHT) {
		if (wrap_method == IGNORE) return 0;
		else if (wrap_method == CLEAR) clearDisplay();
		else if (wrap_method == SCROLL) scroll();
		b = rowcol2mem(row,col);
	}
	if (c=='\n') {
		col=0;
		row++;
	} else if (c=='\r') {
		col=0;
	} else {
		text_buffer[b] = (char)c;
		col++;
		if (col>=LCDTEXTWIDTH) {
			col=0;
			row++;
		}
	}
	return 1;
}

void TinyAdafruit_PCD8544::scroll() {
	uint8_t i;
	for (i = 0; i < LCDTEXTWIDTH*(LCDTEXTHEIGHT-1); i++) {
		text_buffer[i] = text_buffer[i+LCDTEXTWIDTH];
	}
	for (i = LCDTEXTWIDTH*(LCDTEXTHEIGHT-1); i < LCDTEXTWIDTH*LCDTEXTHEIGHT; i++) {
		text_buffer[i] = 0;
	}
	row = LCDTEXTHEIGHT-1;
	col = 0;
}

void TinyAdafruit_PCD8544::setpos(uint8_t new_row, uint8_t new_col) {
	if (new_row < 0 || new_row > LCDTEXTHEIGHT) return;
	if (new_col < 0 || new_col > LCDTEXTWIDTH) return;
	row = new_row;
	col = new_col;
}

void TinyAdafruit_PCD8544::setchar(uint8_t target_row, uint8_t target_col, char s) {
	uint8_t b = rowcol2mem(target_row, target_col);
	if (b > LCDTEXTWIDTH*LCDTEXTHEIGHT) return; // out of buffer
	text_buffer[b] = s;
}

void TinyAdafruit_PCD8544::cs_enable(bool b) { 
	if (_cs > 0) digitalWrite(_cs, !b); 
}

void TinyAdafruit_PCD8544::begin(uint8_t contrast) {
  // set pin directions
  pinMode(_din, OUTPUT);
  pinMode(_sclk, OUTPUT);
  pinMode(_dc, OUTPUT);
  if (_rst > 0)
    pinMode(_rst, OUTPUT);
  if (_cs > 0)
    pinMode(_cs, OUTPUT);

  // toggle RST low to reset
  if (_rst > 0) {
    digitalWrite(_rst, LOW);
    _delay_ms(500);
    digitalWrite(_rst, HIGH);
  }

  clkport     = portOutputRegister(digitalPinToPort(_sclk));
  clkpinmask  = digitalPinToBitMask(_sclk);
  mosiport    = portOutputRegister(digitalPinToPort(_din));
  mosipinmask = digitalPinToBitMask(_din);
  csport    = portOutputRegister(digitalPinToPort(_cs));
  cspinmask = digitalPinToBitMask(_cs);
  dcport    = portOutputRegister(digitalPinToPort(_dc));
  dcpinmask = digitalPinToBitMask(_dc);

  // get into the EXTENDED mode!
  command(PCD8544_FUNCTIONSET | PCD8544_EXTENDEDINSTRUCTION );

  // LCD bias select (4 is optimal?)
  command(PCD8544_SETBIAS | 0x4);

  // set VOP
  if (contrast > 0x7f)
    contrast = 0x7f;

  command( PCD8544_SETVOP | contrast); // Experimentally determined

  // normal mode
  command(PCD8544_FUNCTIONSET);

  // Set display to Normal
  command(PCD8544_DISPLAYCONTROL | PCD8544_DISPLAYNORMAL);

  display();
}

inline void TinyAdafruit_PCD8544::fastSPIwrite(uint8_t d) {
  
  for(uint8_t bit = 0x80; bit; bit >>= 1) {
    *clkport &= ~clkpinmask;
    if(d & bit) *mosiport |=  mosipinmask;
    else        *mosiport &= ~mosipinmask;
    *clkport |=  clkpinmask;
  }
}

inline void TinyAdafruit_PCD8544::slowSPIwrite(uint8_t c) {
  shiftOut(_din, _sclk, MSBFIRST, c);
}

void TinyAdafruit_PCD8544::command(uint8_t c) {
  digitalWrite(_dc, LOW);
  cs_enable(true);
  fastSPIwrite(c);
  cs_enable(false);
}

void TinyAdafruit_PCD8544::data(uint8_t c) {
  digitalWrite(_dc, HIGH);
  cs_enable(true);
  fastSPIwrite(c);
  cs_enable(false);
}

void TinyAdafruit_PCD8544::setContrast(uint8_t val) {
  if (val > 0x7f) {
    val = 0x7f;
  }
  command(PCD8544_FUNCTIONSET | PCD8544_EXTENDEDINSTRUCTION );
  command( PCD8544_SETVOP | val); 
  command(PCD8544_FUNCTIONSET);
  
 }



void TinyAdafruit_PCD8544::display(void) {
	uint8_t col, p;

	for(p = 0; p < 6; p++) {

		command(PCD8544_SETYADDR | p);
		command(PCD8544_SETXADDR | 0);

		digitalWrite(_dc, HIGH); 
		cs_enable(true);
		for(col=0; col < LCDWIDTH; col++) {
			uint8_t ch_idx = col/(FONTWIDTH+1); // which character in this row is it?
			uint8_t ch_col = col%(FONTWIDTH+1); // which pixel column within the character is it?
			unsigned char c = text_buffer[LCDTEXTWIDTH*p + ch_idx]; // get the character from the text buffer
			if (c>0x7F) c=0x7F; // map extended characters to 0x7F
			unsigned char data = (ch_col<FONTWIDTH) ? pgm_read_byte_near(font5x7 + c*FONTWIDTH + ch_col) : 0; // look up this column of pixels from the font for this character
			fastSPIwrite(data);
		}
		cs_enable(false);

	}

	command(PCD8544_SETYADDR );  // no idea why this is necessary but it is to finish the last byte?
}

// clear everything
void TinyAdafruit_PCD8544::clearDisplay(void) {
	uint8_t i;
	for (i=0; i<LCDTEXTWIDTH * LCDTEXTHEIGHT; i++) {
		text_buffer[i] = 0;
	}
	row=0;
	col=0;
}


